;;;;;;;; scheme misc library
;;
;; Windows ManagerpƂC悭ĝƂ
;;
;; Naoki Iwasaki

;;;; list

;; filter
(define (filter comp l)
  (if (null? l) ()
      (let ((cr (car l))
	    (cd (cdr l)))
	(if (comp cr)
	    (cons cr (filter comp cd))
	    (filter comp cd)))))

;; quick sort
(define (sort-by fnc l)
  (if (< (length l) 2) l
      (let* ((p (car l))
	     (t (cdr l))
	     (leq (lambda (d) (fnc d p)))
	     (gt  (lambda (d) (not (leq d)))))
	(append (sort-by fnc (filter leq t))
		(cons p (sort-by fnc (filter gt t)))))))

(define (sort l)
  (sort-by < l))

;; uniq
(define (uniq lst)
  (if (null? lst) ()
      (let ((n (car lst)))
	(cons n (uniq (filter (lambda (x) (not (= n x))) (cdr lst)))))))

(define (nth n l)
  (if (null? l) #f
      (if (< n 0)
	  ()
	  (if (= n 0)
	      (car l)
	      (nth (- n 1) (cdr l))))))

(define (last lst)
  (cond ((null? lst) ())
	((null? (cdr lst)) (car lst))
	(else (last (cdr lst)))))

(define (alist-set l a b)
  (if (null? l) (list (list a b))
      (if (= (caar l) a)
	  (cons (list a b) (cdr l))
	  (cons (car l)    (alist-set (cdr l) a b)))))

(define (alist-find l a)
  (if (null? l) #f
      (if (equal? (caar l) a)
	  (cadar l)
	  (alist-find (cdr l) a))))

(define (show-console flag)
  (if flag 
      (begin (show-window console-hwnd sw-shownormal)
	     (set-foreground-window console-hwnd))
      (show-window console-hwnd sw-hide)))

(define (zip-with proc as bs)
  (if (or (null? as) (null? bs)) ()
      (cons (proc (car as) (car bs)) (zip-with proc (cdr as) (cdr bs)))))

(define (zip as bs)
  (zip-with list as bs))

(define (unzip lst)
  (cond ((null? lst) ())
	((null? (car lst)) ())
	(else (cons (map car lst) (unzip (map cdr lst))))))

(define (foldl f x lst)
  (if (null? lst)
      x
      (foldl f (f x (car lst)) (cdr lst))))

(define (concat a)
  (if (null? a) ()
      (let ((cr (car a)))
	(if (list? cr)
	    (append (concat cr) (concat (cdr a)))
	    (cons cr (concat (cdr a)))))))

(define (inc n) (+ n 1))
(define (dec n) (- n 1))

(define (take n lst)
  (if (null? lst) ()
      (if (<= n 0)
	  ()
	  (cons (car lst) (take (dec n) (cdr lst))))))

(define (replicate n a)
  (if (<= n 0) ()
      (cons a (replicate (dec n) a))))

;; return (front . back)  ;(length front) = n
(define (split-at n lst)
  (if (null? lst) (cons () ())
      (if (<= n 0)
	  (cons () lst)
	  (let ((tmp (split-at (dec n) (cdr lst))))
	    (cons (cons (car lst) (car tmp)) (cdr tmp))))))

;; init.scmfoldŕCfoldl̊ԈႢ̋Ĉŏ㏑
(define (foldr f x lst)
  (if (null? lst)
      x
      (f (car lst) (foldr f x (cdr lst)))))


;;;; for hotkey event
(define events ())

(define (set-key mod vk proc)
  (let ((vk-num (if (char? vk)
		    (char->integer (char-upcase vk))
		    vk)))
    (register-hotkey mod vk-num)
    (set! events (alist-set events (+ mod (* vk-num #x10000)) proc))))

(define (call-event n)
  ((alist-find events n)))


;;;; for create-window window-procedure
(define window-procedures ())

(define (create-window winproc class-name name style x y w h)
  (create-window-ex winproc 0 class-name name style x y w h))

(define (create-window-ex winproc ex-style class-name name style x y w h)
  (let ((hwnd (create-window-ex-imp ex-style class-name name style x y w h)))
    (if (not (null? hwnd))
	(begin
	  (set! window-procedures
		(alist-set window-procedures hwnd winproc))
	  hwnd))))

(define (call-window-procedure hwnd msg wparam lparam)
  ((alist-find window-procedures hwnd) hwnd msg wparam lparam))

(define (destroy-window-procedure hwnd)
  (set! window-procedures (filter (lambda (a) (not (= (car a) hwnd))) window-procedures)))

;;;; Window֌W

;; KWindowHandle炤
(define (to-hwnd v)
  (cond ((number? v) v)
	((string? v) (find-window () v))))
  
(define (get-window-position hwnd)
  (let ((rect (get-window-rect (to-hwnd hwnd))))
    (list (nth 0 rect) (nth 1 rect))))

(define (set-window-position hwnd pos)
  (set-window-pos (to-hwnd hwnd) 0 (car pos) (cadr pos) 0 0 (+ swp-nosize swp-nozorder)))

(define (get-window-size hwnd)
  (let ((rect (get-window-rect (to-hwnd hwnd))))
    (list (nth 2 rect) (nth 3 rect))))

(define (set-window-size hwnd w h)
  (set-window-pos (to-hwnd hwnd) 0 0 0 w h (+ swp-nomove swp-nozorder)))

(define (move pos dx dy)
  (list (+ (car pos) dx) (+ (cadr pos) dy)))

(define (set-nth n lst v)
  (if (null? lst) ()
      (if (= n 0)
	  (cons v (cdr lst))
	  (cons (car lst) (set-nth (dec n) (cdr lst) v)))))

(define (rect-left rect)
  (car    rect))
(define (rect-top rect)
  (cadr   rect))
(define (rect-right rect)
  (caddr  rect))
(define (rect-bottom rect)
  (cadddr rect))
(define (rect-set x rect v)
  (cond ((eq? x 'left  ) (set-nth 0 rect v))
	((eq? x 'top   ) (set-nth 1 rect v))
	((eq? x 'right ) (set-nth 2 rect v))
	((eq? x 'bottom) (set-nth 3 rect v))
	(else (error "rect-set need 'left, 'top, 'right or 'bottom"))))
(define (rect-get x rect)
  (cond ((eq? x 'left  ) (rect-left   rect))
	((eq? x 'top   ) (rect-top    rect))
	((eq? x 'right ) (rect-right  rect))
	((eq? x 'bottom) (rect-bottom rect))
	(else
	 (error "rect-get need 'left, 'top, 'right or 'bottom"))))
  
(define (rect-width rect)
  (- (rect-right rect) (rect-left rect)))
(define (rect-height rect)
  (- (rect-bottom rect) (rect-top rect)))

;;;; lambdâɔꂽ

;; curry (map (curry + 1) '(0 1 2)) => (1 2 3)
(define (curry fn . args)
  (lambda more-args (apply fn (append args more-args))))

;; ֐ ((compose inc inc) 1) => (inc (inc 1))
(define (compose f g)
  (lambda x (f (apply g x))))

;;;; ̑

;;ɂ鎞Ԃ̌v
(define-macro (time a)
  (time-begin-period 1)
  (define t (time-get-time))
  (eval a)
  (set! t (- (time-get-time) t))
  (time-end-period 1)
  t)

(define pi (acos -1))
