/** SchemepWindowsAPI
 */

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <windowsx.h>
#include <process.h>
#include <mmsystem.h>
#include <shellapi.h>

#include <string>
#include <map>

using namespace std;

#pragma comment(lib, "winmm.lib")

#include "scheme-private.h"

extern HINSTANCE hInst;
extern HWND hMainWnd, hCWnd;
extern CRITICAL_SECTION schemeMutex;

scheme*pScheme;

// defaultWindowProcedureւ̃}bv
map<HWND, WNDPROC> defaultWinProcMap;

typedef pointer schemeFunction(scheme*, pointer);

template<typename T>
void arrayToList(scheme*sc, pointer&list, T array[], int length, pointer (*v2p)(scheme*,T))
{
    list=sc->NIL;
    while(length>0)
        list=cons(sc, v2p(sc, array[--length]), list);
}


// XgłȂ̌^`FbN
bool checkType(scheme*sc, pointer value, char type)
{
    switch(type)
    {
    case 'd':
        return is_number(value)!=0;
    case 'r':
        return is_real(value)!=0;
    case 'c':
        return is_character(value)!=0;
    case 's':
        return is_string(value)!=0;
    case 'l':
        return is_closure(value)!=0;
    default:
        return false; // unknown type
    }
}

bool checkTypeImp(scheme*sc, pointer value, string&type)
{
    /*
    typeȂ`FbN
    */
    if(type.empty())
        return false;

    if(type[0]=='[')
    {
        if(!is_pair(value))
            return false;
        type.erase(0,1);

        while(type[0]!=']')
        {
            if(!is_pair(value))
                return false;

            if(!checkTypeImp(sc, pair_car(value), type))
                return false;
            value=pair_cdr(value);
        }
        if(sc->NIL != value)
            return false;
    }
    else if(type[0]=='(')
    {
        if(!is_pair(value))
            return false;

        type.erase(0,1);

        if(!checkTypeImp(sc, pair_car(value), type))
            return false;

        if(!checkTypeImp(sc, pair_cdr(value), type))
            return false;

        if(type[0]!=')')
            return false;
    }
    else// if(!isspace(type[0])) Xy[XƁCpairƂlistȂƂȂD(߂ǂ)
    {
        if(!checkType(sc, value, type[0]))
            return false;
    }
    type.erase(0,1);
    return true;
}

/** schemě^ςȕŃ`FbN
    d -> integer
    r -> real
    c -> char
    s -> string
    l -> closure(lambda)
    [] -> list
    (ab) -> pair

    [dddc(sd)] == '(1 1 1 #\a ("hoge" . 1)) Ȋ
 */
bool checkType(scheme*sc, pointer value, const char*type)
{
    string t(type);
    bool ret=checkTypeImp(sc, value, t);

    return ret && t.empty();
}

/// ֐̌^`FbN+G[bZ[Wf
bool checkFuncType(scheme*sc, pointer args, const char*func, const char*type)
{
    if(!checkType(sc, args, type))
    {
        fprintf(stderr, "Error: %s: arguments must be: %s\n", func, type);
        return false;
    }
    return true;
}

pointer scmLogand(scheme*sc, pointer args)
{
    unsigned int ret= ~0;
    while(is_pair(args))
    {
        ret&=ivalue(pair_car(args));
        args=pair_cdr(args);
    }
    return mk_integer(sc, ret);
}

pointer nth(pointer list, int n)
{
    pointer p=list;
    while(n>0)
    {
        p=pair_cdr(p);
        --n;
    }
    return pair_car(p);
}

void registerHotKey(UINT mod, UINT vk); // @main.cpp ^^;
pointer scmRegisterHotKey(scheme*sc, pointer args)
{
    UINT mod =ivalue(nth(args, 0));
    pointer pvk = nth(args, 1);
    UINT vk = is_character(pvk)?
        toupper(charvalue(pvk))
        :ivalue(pvk);

    // WindowsAPIRegisterHotKeyĂяoĂ͂ȂD
    // APÍCĂяõXbhɃzbgL[o^̂Ȃ̂
    // R\[ォ炱APIĂ񂾏ꍇɁCWinProc̃[vĂ
    // XbhCzbgL[EƂłȂD
    // registerHotKeýCׂ̊֐
    registerHotKey(mod, vk);
    return sc->T;
}

pointer scmGetForeGroundWindow(scheme*sc, pointer args)
{
    HWND h=GetForegroundWindow();
    long l=(long)(long long)h;
    return mk_integer(sc, l);
}

pointer scmGetWindowRect(scheme*sc, pointer args)
{
    HWND hWnd=(HWND)(long long)ivalue(nth(args, 0));
    RECT winRect;
    if(!GetWindowRect(hWnd, &winRect))
        return sc->NIL;
    
    long ar[4]={winRect.left, winRect.top, winRect.right, winRect.bottom};

    pointer ret;
    arrayToList<long>(sc, ret, ar, 4, mk_integer);
    return ret;
}

pointer scmSetWindowRect(scheme*sc, pointer args)
{
    HWND hWnd    = (HWND)(long long)ivalue(nth(args, 0));
    pointer rect = nth(args, 1);
    int x =ivalue(nth(rect, 0)),
        y =ivalue(nth(rect, 1)),
        x2=ivalue(nth(rect, 2)),
        y2=ivalue(nth(rect, 3));

    return mk_integer(sc, MoveWindow(hWnd, x, y, x2-x, y2-y, true));
}

pointer scmSetWindowPos(scheme*sc, pointer args)
{
    HWND hWnd            = (HWND)(long long)ivalue(nth(args, 0));
    HWND hWndInsertAfter = (HWND)(long long)ivalue(nth(args, 1));
    int x                = ivalue(nth(args, 2)),
        y                = ivalue(nth(args, 3)),
        w                = ivalue(nth(args, 4)),
        h                = ivalue(nth(args, 5));
    UINT uFlags          = ivalue(nth(args, 6));

    return mk_integer(sc,
        SetWindowPos(hWnd, hWndInsertAfter, x, y, w, h, uFlags));
}

// (show-window hwnd flag)
pointer scmShowWindow(scheme*sc, pointer args)
{
    HWND hWnd=(HWND)(long long)ivalue(nth(args, 0));
    int flag=ivalue(nth(args, 1));
    return mk_integer(sc, ShowWindow(hWnd, flag));
}

pointer scmSetActiveWindow(scheme*sc, pointer args)
{
    HWND hWnd=(HWND)(long long)ivalue(nth(args, 0));
    return mk_integer(sc, (long)(long long)SetActiveWindow(hWnd));
}

// IɃtHAOEh
BOOL SetAbsForegroundWindow(HWND hWnd)
{
	int nTargetID, nForegroundID;
	DWORD sp_time;
	nForegroundID = GetWindowThreadProcessId(GetForegroundWindow(), NULL);
	nTargetID = GetWindowThreadProcessId(hWnd, NULL );
	AttachThreadInput(nTargetID, nForegroundID, TRUE );  // Xbhѕt
	SystemParametersInfo( SPI_GETFOREGROUNDLOCKTIMEOUT,0,&sp_time,0);
	SystemParametersInfo( SPI_SETFOREGROUNDLOCKTIMEOUT,0,(LPVOID)0,0);
	BOOL ret = SetForegroundWindow(hWnd);
	SystemParametersInfo( SPI_SETFOREGROUNDLOCKTIMEOUT,0,(LPVOID)(long long)sp_time,0);
	AttachThreadInput(nTargetID, nForegroundID, FALSE );  // Xbh؂藣

    return ret;
}

pointer scmSetForegroundWindow(scheme*sc, pointer args)
{
    HWND hWnd=(HWND)(long long)ivalue(nth(args, 0));
    return mk_integer(sc, (long)(long long)SetAbsForegroundWindow(hWnd));
}

// (find-window className windowName)
pointer scmFindWindow(scheme*sc, pointer args)
{
    pointer className  = nth(args, 0);
    pointer windowName = nth(args, 1);

    char*classNameStr;
    char*windowNameStr;

    classNameStr =is_string(className) ?string_value(className) :NULL;
    windowNameStr=is_string(windowName)?string_value(windowName):NULL;

    return mk_integer(sc, (long)(long long)FindWindow(classNameStr, windowNameStr));
}

pointer scmGetWindow(scheme*sc, pointer args)
{
    HWND hWnd=(HWND)(long long)ivalue(nth(args, 0));
    UINT uCmd=ivalue(nth(args, 1));
    return mk_integer(sc, (long)(long long)GetWindow(hWnd, uCmd));
}
pointer scmGetNextWindow(scheme*sc, pointer args)
{
    HWND hWnd=(HWND)(long long)ivalue(nth(args, 0));
    UINT wCmd=ivalue(nth(args, 1));
    return mk_integer(sc, (long)(long long)GetNextWindow(hWnd, wCmd));
}

pointer scmGetWindowStyle(scheme*sc, pointer args)
{
    HWND hWnd=(HWND)(long long)ivalue(nth(args, 0));
    return mk_integer(sc, GetWindowStyle(hWnd));
}

pointer scmGetWorkarea(scheme*sc, pointer args)
{
    RECT deskRect;
    SystemParametersInfo(SPI_GETWORKAREA, 0, &deskRect, 0);

    long ar[4]={deskRect.left, deskRect.top, deskRect.right, deskRect.bottom};

    pointer ret;
    arrayToList<long>(sc, ret, ar, 4, mk_integer);
    return ret;
}

struct EnumWindowsParam
{
    scheme*sc;
    pointer*pWinList;
    HWND hTaskWnd;
};

BOOL CALLBACK enumWindowsProc(HWND hWnd, LPARAM lParam)
{
    EnumWindowsParam*pParam=(EnumWindowsParam*)lParam;
    scheme*sc=pParam->sc;

    *pParam->pWinList=_cons(sc, mk_integer(sc, (long)(long long)hWnd), *pParam->pWinList, 0);
    return TRUE;
}

// (enum-windows)
// WindowHandlẽXgԂ
pointer scmEnumWindows(scheme*sc, pointer args)
{
    pointer winList=sc->NIL;
    EnumWindowsParam param={sc, &winList};
 
    EnumDesktopWindows(NULL, enumWindowsProc, (LPARAM)&param);
    //EnumWindows(enumWindowsProc, (LPARAM)&param);
    return winList;
}

BOOL CALLBACK enumVisibleWindowsProc(HWND hWnd, LPARAM lParam)
{
    int style=style = GetWindowStyle(hWnd);
    if(   !IsWindow(hWnd)
       || !IsWindowVisible(hWnd)
       || !((style & WS_CAPTION)) )// && !(style & WS_POPUP)))
       return TRUE; // 񋓂WindowɒǉȂ

    EnumWindowsParam*pParam=(EnumWindowsParam*)lParam;
    if(pParam->hTaskWnd==hWnd)
        return TRUE; // 񋓂WindowɒǉȂ

    scheme*sc=pParam->sc;
    *pParam->pWinList=_cons(sc, mk_integer(sc, (long)(long long)hWnd), *pParam->pWinList, 0);
    return TRUE;
}

// (enum-visible-windows)
// ĂWindowHandlẽXgԂ
pointer scmEnumVisibleWindows(scheme*sc, pointer args)
{
    pointer winList=sc->NIL;
    EnumWindowsParam param={sc, &winList, FindWindow("Shell_TrayWnd",NULL)};
 
    EnumDesktopWindows(NULL, enumVisibleWindowsProc, (LPARAM)&param);
    //EnumWindows(enumWindowsProc, (LPARAM)&param);
    return winList;
}

pointer scmIsWindow(scheme*sc, pointer args)
{
    HWND hWnd=(HWND)(long long)ivalue(nth(args, 0));
    return IsWindow(hWnd)?sc->T:sc->F;
}

pointer scmIsWindowVisible(scheme*sc, pointer args)
{
    HWND hWnd=(HWND)(long long)ivalue(nth(args, 0));
    return IsWindowVisible(hWnd)?sc->T:sc->F;
}
pointer scmIsIconic(scheme*sc, pointer args)
{
    HWND hWnd=(HWND)(long long)ivalue(nth(args, 0));
    return IsIconic(hWnd)?sc->T:sc->F;
}
pointer scmIsZoomed(scheme*sc, pointer args)
{
    HWND hWnd=(HWND)(long long)ivalue(nth(args, 0));
    return IsZoomed(hWnd)?sc->T:sc->F;
}

pointer scmGetWindowText(scheme*sc, pointer args)
{
    HWND hWnd=(HWND)(long long)ivalue(nth(args, 0));
    char text[100];
    GetWindowText(hWnd, text, 100);
    return mk_string(sc, text);
}

pointer scmGetTickCount(scheme*sc, pointer args)
{
    return mk_integer(sc, GetTickCount());
}

pointer scmTimeGetTime(scheme*sc, pointer args)
{
    return mk_integer(sc, timeGetTime());
}

pointer scmTimeBeginPeriod(scheme*sc, pointer args)
{
    return mk_integer(sc, timeBeginPeriod(ivalue(pair_car(args))));
}

pointer scmTimeEndPeriod(scheme*sc, pointer args)
{
    return mk_integer(sc, timeEndPeriod(ivalue(pair_car(args))));
}

pointer scmSendMessage(scheme*sc, pointer args)
{
    HWND   hWnd  =(HWND)(long long)ivalue(nth(args, 0));
    UINT   msg   =                 ivalue(nth(args, 1));
    WPARAM wParam=(WPARAM)         ivalue(nth(args, 2));
    LPARAM lParam=(LPARAM)         ivalue(nth(args, 3));

    return mk_integer(sc, (long)(long long)SendMessage(hWnd, msg, wParam, lParam));
}

pointer scmSleep(scheme*sc, pointer args)
{
    DWORD ms = ivalue(pair_car(args));
    Sleep(ms);
    return sc->NIL;
}

pointer scmLaunch(scheme*sc, pointer args)
{
    pointer pcmd = nth(args, 0);
    if(!is_string(pcmd))
        return sc->F;
    char*cmd = string_value(pcmd);

    char*cd;
    if(pair_cdr(args)==sc->NIL)
        cd=NULL;
    else
    {
        pointer pcd=nth(args, 1);
        if(!is_string(pcd))
            return sc->F;
        cd=string_value(pcd);
    }

    PROCESS_INFORMATION pi;
	STARTUPINFO si;
	ZeroMemory(&pi,sizeof(pi));
	ZeroMemory(&si,sizeof(si));

	CreateProcess(NULL, cmd, NULL, NULL, FALSE, CREATE_NEW_CONSOLE, NULL, cd, &si, &pi);

    return sc->T;
}

pointer scmDestoryWindow(scheme*sc, pointer args)
{
    // hwnd
    if(!checkFuncType(sc, args, "create-window-ex", "[d]"))
        return sc->NIL;

    HWND hwnd = (HWND)(long long)ivalue(nth(args, 0));

    DestroyWindow(hwnd);
    return sc->NIL;
}

struct WindowData
{
    scheme*sc;
    DWORD exStyle;
    char*className;
    char*name;
    DWORD style;
    int x,
        y,
        w,
        h;

    // l̂p
    bool evaled;
    HWND hwnd;
};


LRESULT CALLBACK windowProcedure(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
    /*
    if(msg==WM_GETTEXT)
    {
        int a;
        a=1;
    }
    if(msg==WM_KEYDOWN)
    {
        int a;
        a=1;
    }
    */
    printf("count : %d\n", schemeMutex.RecursionCount);
    if(schemeMutex.RecursionCount==0)
    {
        // schemegȂȂCdefaultɂ܂ႤDDD
        EnterCriticalSection(&schemeMutex);
        pointer args=cons(pScheme, mk_integer(pScheme, (long)(long long)hWnd)
            , cons(pScheme, mk_integer(pScheme, msg)
            , cons(pScheme, mk_integer(pScheme, (long)wParam)
            , cons(pScheme, mk_integer(pScheme, (long)lParam), pScheme->NIL))));

        scheme_apply1(pScheme, "call-window-procedure", args);
        LeaveCriticalSection(&schemeMutex);
    }

    WNDPROC proc = defaultWinProcMap.find(hWnd)->second;
    return CallWindowProc(proc, hWnd, msg, wParam, lParam);
}

void windowLoop(void*pdata)
{
    WindowData*wd=(WindowData*)pdata;

    scheme*sc=wd->sc;

    HWND hWnd = CreateWindowEx(
        wd->exStyle,
        wd->className,
        wd->name,
        wd->style,
        wd->x,wd->y,wd->w,wd->h,
        NULL, // parent
        NULL,     // menu
        hInst,
        0);       // lParam
    ShowWindow(hWnd, SW_SHOW);

    WNDPROC defaultWinProc = (WNDPROC)(long long)GetWindowLong(hWnd, GWL_WNDPROC);
	SetWindowLong(hWnd, GWL_WNDPROC, (LONG)(long long)windowProcedure);
    defaultWinProcMap.insert(make_pair(hWnd, defaultWinProc));

    wd->hwnd=hWnd;
    wd->evaled=true;

    MSG msg;
    do
    {
        //if(PeekMessage(&msg, NULL, 0, 0, PM_REMOVE))
        if(GetMessage(&msg, NULL, 0, 0))
        {
            TranslateMessage(&msg);
            DispatchMessage(&msg);
        }
    }while(msg.message!=WM_QUIT);

    EnterCriticalSection(&schemeMutex);
    {
        pointer args=cons(sc, mk_integer(sc, (long)(long long)hWnd), sc->NIL);
        scheme_apply1(sc, "destroy-window-procedure", args);
    }
    LeaveCriticalSection(&schemeMutex);

    return;
}

pointer scmCreateWindowEx(scheme*sc, pointer args)
{
    // exStyle className name style x y w h
    if(!checkFuncType(sc, args, "create-window-ex", "[dssddddd]"))
        return sc->NIL;

    WindowData*wd=(WindowData*)malloc(sizeof(WindowData));

    wd->sc=sc;
    wd->exStyle  = ivalue      (nth(args, 0));
    wd->className = string_value(nth(args, 1));
    wd->name      = string_value(nth(args, 2));
    wd->style    = ivalue      (nth(args, 3));
    wd->x          = ivalue      (nth(args, 4));
    wd->y          = ivalue      (nth(args, 5));

    wd->w          = ivalue      (nth(args, 6));
    wd->h          = ivalue      (nth(args, 7));

    wd->evaled=false;
    wd->hwnd=0;

    _beginthread(windowLoop, 0, (void*)wd);

    // œK}
    volatile WindowData*vwd=wd;
    while(!vwd->evaled)
        ;
    HWND hwnd=wd->hwnd;

    free(wd);

    return mk_integer(sc, (long)(long long)hwnd);
}

pointer scmGetSystemMetrics(scheme*sc, pointer args)
{
    if(!checkFuncType(sc, args, "get-system-metrics", "[d]"))
        return sc->NIL;

    return mk_integer(sc, GetSystemMetrics(ivalue(pair_car(args))));
}

pointer scmPostQuitMessage(scheme*sc, pointer args)
{
    if(!checkFuncType(sc, args, "pos-quit-message", "[d]"))
        return sc->NIL;

    PostQuitMessage(ivalue(pair_car(args)));
    return sc->NIL;
}

void registeAPIs(scheme*sc)
{
    pScheme=sc;

#   define DEF_FF(name, func) scheme_define(sc, sc->global_env,\
        mk_symbol(sc, name),\
        mk_foreign_func(sc, func))

    DEF_FF("logand"      , scmLogand);
    DEF_FF("register-hotkey"      , scmRegisterHotKey);
    DEF_FF("set-window-pos", scmSetWindowPos);
    DEF_FF("set-window-rect", scmSetWindowRect);
    DEF_FF("get-window-rect", scmGetWindowRect);
    DEF_FF("get-workarea", scmGetWorkarea);
    DEF_FF("show-window", scmShowWindow);
    DEF_FF("set-active-window", scmSetActiveWindow);
    DEF_FF("set-foreground-window", scmSetForegroundWindow);
    DEF_FF("get-foreground-window", scmGetForeGroundWindow);
    DEF_FF("find-window", scmFindWindow);
    DEF_FF("get-window", scmGetWindow);
    DEF_FF("get-window-style", scmGetWindowStyle);
    DEF_FF("get-next-window", scmGetNextWindow);
    DEF_FF("enum-windows", scmEnumWindows);
    DEF_FF("enum-visible-windows", scmEnumVisibleWindows);
    DEF_FF("is-window", scmIsWindow);
    DEF_FF("is-window-visible", scmIsWindowVisible);
    DEF_FF("is-iconic", scmIsIconic);
    DEF_FF("is-zoomed", scmIsZoomed);
    DEF_FF("get-window-text", scmGetWindowText);
    DEF_FF("get-tick-count", scmGetTickCount);
    DEF_FF("time-get-time", scmTimeGetTime);
    DEF_FF("time-begin-period", scmTimeBeginPeriod);
    DEF_FF("time-end-period", scmTimeEndPeriod);
    DEF_FF("send-message", scmSendMessage);
    DEF_FF("sleep", scmSleep);
    DEF_FF("launch", scmLaunch);
    DEF_FF("create-window-ex-imp", scmCreateWindowEx);
    DEF_FF("destroy-window", scmDestoryWindow);
    DEF_FF("get-system-metrics", scmGetSystemMetrics);
    DEF_FF("post-quit-message", scmPostQuitMessage);
#   undef DEF_FF
}
