/** WindowsManager

    Naoki Iwasaki(safii@flightless-wing.com)
*/

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <process.h>

#include <vector>
#include <map>
#include <string>
#include <iostream>
#include <algorithm>

#include "scheme.h"
#include "scheme-private.h"
#include "api.h"

using namespace std;

const char appName[]="Windows Manager";

scheme*sc=NULL;
bool createdWindow=false;
HWND hMainWnd=NULL, hCWnd=NULL;
string inputBuf;
HINSTANCE hInst;

CRITICAL_SECTION hotkeyMutex;
CRITICAL_SECTION schemeMutex;
int hotKeyId=0;

vector<pair<UINT, UINT> > regHotkeys;

/** WinProcĂXbhȊOzbgL[o^
*/
void registerHotKey(UINT mod, UINT vk)
{
    EnterCriticalSection(&hotkeyMutex);
    {
        regHotkeys.push_back(make_pair(mod, vk));
    }
    LeaveCriticalSection(&hotkeyMutex);
}

void RegisterHotKeyPair(pair<UINT, UINT> mod_vk)
{
    RegisterHotKey(hMainWnd, hotKeyId++, mod_vk.first, mod_vk.second);
}
void checkRegisterHotKey()
{
    EnterCriticalSection(&hotkeyMutex);
    {
        for_each(regHotkeys.begin(), regHotkeys.end(), RegisterHotKeyPair);
        regHotkeys.clear();
    }
    LeaveCriticalSection(&hotkeyMutex);
}

bool loadSchemeFile(const char*fileName)
{

  //init.scm̓ǂݍ
  FILE *fp=fopen(fileName, "r");
  if(!fp)
      return false;

  scheme_load_file(sc,fp);
  fclose(fp);
  return true;
}

void registeForeighData()
{
#   define DEF_FF(name, func) scheme_define(sc, sc->global_env,\
        mk_symbol(sc, name),\
        mk_foreign_func(sc, func))

#   undef DEF_FF

    registeAPIs(sc);
}

bool initScheme()
{
    if(sc)
        scheme_deinit(sc);
    sc=scheme_init_new();

    if(!scheme_init(sc))
        return false;

    scheme_set_input_port_file(sc, stdin);
    scheme_set_output_port_file(sc, stdout);

    // hwnddefine
    scheme_define(sc, sc->global_env,
            mk_symbol(sc, "my-hwnd"),
                mk_integer(sc, (long)(long long)hMainWnd));

    while(!hCWnd)
        Sleep(10);
    scheme_define(sc, sc->global_env,
            mk_symbol(sc, "console-hwnd"),
                mk_integer(sc, (long)(long long)hCWnd));

    registeForeighData();

    loadSchemeFile("system.scm");

    return true;
}

LRESULT CALLBACK winProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
    switch(msg)
    {
    case WM_DESTROY:
        PostQuitMessage(0);
        break;
    case WM_HOTKEY:
        EnterCriticalSection(&schemeMutex);
        {
            pointer args=cons(sc, mk_integer(sc, (unsigned long)lParam), sc->NIL);
            scheme_apply1(sc, "call-event", args);
        }
        LeaveCriticalSection(&schemeMutex);
        break;
    }
    return TRUE;
}

int exprParentDepth(const string expr)
{
    int depth=0;
    string::const_iterator it, end=expr.end();
    for(it=expr.begin(); it!=end; it++)
    {
        if(*it=='(')
            depth++;
        else if(*it==')')
            depth--;

        if(depth<0)
            return -1; // ŁC[0ɂȂ܂
    }

    return depth;
}

/** ]
    
    @retval true  ]
    @retval false ]Ȃ
*/
bool evalLisp(const string&expr)
{
    if(expr.empty())
        return false;

    inputBuf+=expr+"\n";

    int depth=exprParentDepth(inputBuf);

    if(depth<0)
    {
        cerr << "error : too many right parent." << endl;
        inputBuf.clear();
        return true;
    }

    if(depth>0)
        return false;

    scheme_load_string(sc, inputBuf.c_str());
    inputBuf.clear();

    int retcode=sc->retcode;
    std::string ret=scheme_toString(sc, sc->value);

    if(is_proc(sc->value) && ivalue(sc->value)==140)
        throw 0;

    // ]
    if(retcode==0 && !ret.empty())
        cout << ret << endl;

    return true;
}

// s̏ԂAllocConsole
void hideAllocConsole()
{
	DWORD*	pflags = (DWORD*) 0x00020068;
	WORD*	pshow  = (WORD*)  0x0002006C;

	DWORD	backup_flags = *pflags;
	WORD	backup_show  = *pshow;

	STARTUPINFO si;
	GetStartupInfo(&si);

	if(si.dwFlags == backup_flags && si.wShowWindow == backup_show)
    {
        *pflags |= STARTF_USESHOWWINDOW;
        *pshow  = SW_HIDE;
	}

	AllocConsole();

	*pflags = backup_flags;
	*pshow  = backup_show;
}

void consoleLoop(void*threadID)
{
    //R\[쐬CWo͂Ɋ蓖Ă
    /*
    hideAllocConsole();
	/*/
    AllocConsole();
    //*/

    SetConsoleTitle(appName);
    while(!(hCWnd=FindWindow(NULL, appName)))
        Sleep(10);

	freopen("CONOUT$", "w", stdout);
	freopen("CONOUT$", "w", stderr);
	freopen("CONIN$", "r", stdin);

    string buf;
    bool evaled=true;
    while(1)
    {
        if(evaled)
            cout << "wmscm> ";
        else
            cout << "=====> ";

        getline(cin, buf);
        // lockǂǂĂȂ
        try{
            EnterCriticalSection(&schemeMutex);
            {
                evaled = evalLisp(buf);
            }
            LeaveCriticalSection(&schemeMutex);
        }
        catch(...)
        {
            return;
        }
    }
}

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance,
	LPSTR lpCmdLine, int CmdShow)
{
	WNDCLASSEX wcex;
	wcex.cbSize			= sizeof(WNDCLASSEX);
	wcex.style			= 0;
	wcex.lpfnWndProc	= winProc;
	wcex.cbClsExtra		= 0;
	wcex.cbWndExtra		= 0;
	wcex.hInstance		= hInstance;
	wcex.hIcon			= LoadIcon(NULL, IDI_APPLICATION);
	wcex.hCursor		= LoadCursor(NULL, IDC_ARROW);
	wcex.hbrBackground	= (HBRUSH)GetStockObject(LTGRAY_BRUSH);
	wcex.lpszMenuName	= NULL;
	wcex.lpszClassName	= appName;
	wcex.hIconSm		= NULL;
	
    hInst=hInstance;

    if(!RegisterClassEx(&wcex))
    {
        MessageBox(NULL, "WindowClass̓o^s", "G[", MB_ICONWARNING);
        return 0;
    }

    if(!(hMainWnd = CreateWindowEx(0, appName, appName,
        WS_POPUP, CW_USEDEFAULT, 0, CW_USEDEFAULT, 0, NULL, NULL, hInstance, NULL)))
    {
        MessageBox(NULL, "Window̍쐬s", "G[", MB_ICONWARNING);
        return 0;
    }
    createdWindow=true;

    InitializeCriticalSection(&hotkeyMutex);
    InitializeCriticalSection(&schemeMutex);

    HANDLE hThread=(HANDLE)_beginthread(consoleLoop, 0, NULL);

    initScheme();

    MSG msg;
    do
    {
        if(PeekMessage(&msg, NULL, 0, 0, PM_REMOVE))
        {
            TranslateMessage(&msg);
            DispatchMessage(&msg);
        }
        else
            checkRegisterHotKey();

        DWORD ret=WaitForSingleObject(hThread, 1);
        if(ret==0 || ret==-1)
            DestroyWindow(hMainWnd);
    }while(msg.message!=WM_QUIT);

    scheme_deinit(sc);

    DeleteCriticalSection(&hotkeyMutex);
	return  0;
}
